#include <pic.h>				// pic specific identifiers
#define _XTAL_FREQ  4000000		// Xtal speed
__CONFIG(0x3D18);				// Config bits

#define LCD_PORT	PORTA			// LCD configuration
#define LCD_TRIS	TRISA
#define LCD_RS		RA4
#define LCD_RW		RA6
#define LCD_E		RA7

#define I2C_PORT	PORTB			// I2C port configuration
#define I2C_TRIS	TRISB
#define SDAI		RB7
#define SCLI		RB6
#define SDA			TRISB7		// in C  we can control the input / output
#define SCL			TRISB6		// by one simple command.

#define ErrFlag		0x0				// errors for
#define StartFlag	0x1				// those who wish to use them
#define One			0x2
#define Zero		0x3

#define CHIP_Write	0xA0		// standard eeprom address fro writing
#define CHIP_Read	0xA1		// and reading

		// Required prototypes

void LCD_init(void), LCD_cmd(unsigned char ch), LCD_busy(void);
void LCD_printC(const char *str), LCD_printR(char *str);
void LCD_goto(char line, char column), LCD_clr(void);
void LCD_cur(unsigned char ch), pulse_E(void), LCD_hex(int value);
void LCD_char(unsigned char ch), LCD_charD(unsigned char ch);
void I2C_Sendnack(void), I2C_ack(void), I2C_Stop(void), clock(void);
void I2C_Start(void), I2C_Write(unsigned char ch);
char I2C_nack(), E_Write(int addr, unsigned char ch);
unsigned char I2C_Read(void), E_Read(int addr, char* ch);

unsigned char HEX_Table[] = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36,
							0x37, 0x38, 0x39, 0x41, 0x42, 0x43,
							0x44, 0x45, 0x46};

char ErrFlags;			// Read error. (I don't use it, but you might!)
char lcdBuffer[17];		// a screen buffer bi enough for 16 chars + terminator
// The SDA and SCL pins are ALWAYS set at 0 we use tristated pins to
// control the control.. while SDA and SCL are inputs the external pullup
// resistors give a high conditon whilst when SDA or SCL are outputs
// the control lines are pulled low.

void main(void)							// program entry
	{
	int index = 0;						// text variable
	unsigned char ch;
	OPTION = 0x88;						// timer pre-scaler
	CMCON = 0x7;						// Comparitors off
	LCD_TRIS = 0b00000000;				// Led port as outputs
	I2C_TRIS = 0b11111111;				// Joy stick port
	__delay_ms(150);
	RA6 = 0;							// RW low
	RB6 = 0;
	RB7 = 0;
	I2C_PORT = 0;
	LCD_init();							// Initalise screen
	LCD_goto(1,0);
	LCD_printC("Writing..");
	LCD_goto(2,0);
	for(index = 65; index <89; index++)  // A through X
		{
		E_Write(index - 65 , index);
		__delay_ms(5);					// Remember the write delay!!
		}
	LCD_goto(1,0);
	LCD_printC("Reading..");
	for(index = 0; index < 24; index+=4)
		{
		LCD_goto(2,0);
		E_Read(index,&ch);
		LCD_char(ch);
		LCD_char(' ');
		E_Read(index+1,&ch);
		LCD_char(ch);
		LCD_char(' ');
		E_Read(index+2,&ch);
		LCD_char(ch);
		LCD_char(' ');
		E_Read(index+3,&ch);
		LCD_char(ch);
		__delay_ms(180);
		__delay_ms(180);
		}
	LCD_goto(1,0);
	LCD_printC("Finished.");
	LCD_goto(2,0);
	LCD_printC("           ");
	while(1)							// endless Loop
		{
		}
	}

// Write to eeprom this returns a 0 if it fails or a 1 if all gooes well
// For byte eeproms and RTC devices.. comment out the Hi address byte
// and the I2C_nack() that follows it in both these functions.

// This is random writing. Write to a specified address

char E_Write(int addr, unsigned char ch)
	{
	I2C_Start();									// Send a start condition
	I2C_Write(CHIP_Write);							// chip write address
	if(!I2C_nack()) return 0;						// wait for ack
	I2C_Write((unsigned char) (addr >> 8 & 0xff));	// Comment out if single byte requires
	if(!I2C_nack()) return 0;						// wait for ack
	I2C_Write((unsigned char) addr & 0xff);			// low address
	if(!I2C_nack()) return 0;						// wait for ack
	I2C_Write(ch);									// Write values to Eeprom
	if(!I2C_nack()) return 0;
	I2C_Stop();										// Send a stop.
	return 1;										// All went well
	}

// Returns a 1 for error and 0 for OK.
// The function takes an address and a pointer to a character

// This is random reading. Read from a specified addess

unsigned char E_Read(int addr, unsigned char *ch)
	{
	unsigned char byte;
	ErrFlags = 0;									// Clear error
	I2C_Start();
	I2C_Write(CHIP_Write);
	if(!I2C_nack()) return  1;						// Whoops!! problems
	I2C_Write((unsigned char)( addr >> 8 & 0xff));	// Comment out if single byte requires
	if(!I2C_nack()) return  1;
	I2C_Write((unsigned char) addr & 0xff);			//
	if(!I2C_nack()) return 1;
	I2C_Start();
	I2C_Write(CHIP_Read);							//
	if(!I2C_nack()) return  1;
	*ch = I2C_Read();								// read character
	I2C_Sendnack();
	I2C_Stop();
	return 0;
	}

void I2C_Start()		//  start combo
	{
	SDA = 1;
	SCL = 1;
	SDA = 0;			// bring SDA low while SCL is high
	SCL = 0;
	}

void I2C_Stop()			// stop combo
	{
	SCL = 0;
	SDA = 0;
	SCL = 1;
	SDA = 1;			// bring SDA high while SCL is high
	}

unsigned char I2C_Read()
	{
	char index;
	unsigned char byte = 0;						// initialize in byte
	for(index = 0x80; index > 0 ;index>>=1)		// using index as a mask
		{										// cycles through bit 7 to bit 0
		if(SDAI) byte += index;					// read in the SDA line
		clock();
		}
	return byte;								// 8 bits in...
	}

void I2C_Write(unsigned char ch)
	{
	char index;

	for(index = 0x80; index > 0 ;index>>=1)		// using index as mask
		{
		if(ch & index) 							// bit high or low
			SDA = 1;
		else
			SDA = 0;
		clock();
		}
	}

void I2C_ack()							// generate an ack
	{
	SDA = 0;
	clock();
	SDA = 1;
	}

void I2C_Sendnack()						// generate a not ack
	{
	SDA = 1;
	clock();
	}

char I2C_nack()							// recive a not ack
	{
	unsigned char timeout = 255;
	SDA = 1;
	SCL = 1;
	while(SDAI)							// wait till SDA goes low
		{
		if(timeout-- == 0) return 0;	// whoops!! no ack recieved
		}
	SCL = 0;
	return 1;
	}

void clock()							// Might work without the nop
	{
	SCL=1;
	asm("nop");
	SCL=0;
	}

// In C we can use Pointers, Pointers are advanced programming
// and can tie you in knots. The simplest pointer to use is the
// array pointer using "rom char" allows the pointer to
// point to constant character in code section.

void LCD_printC(const char * str)	// This passes the start a ROM character array
	{								// by default the pointer points to data section
	while(*str != 0)				// while the character pointed to isn't 0
		LCD_char(*str++);			// print out the character, then increment
	}								// the pointer down the array

void LCD_printR(char * str)	 		// This passes the start of a RAM character array
	{								// by default the pointer points to data section
	while(*str != 0)				// while the character pointed to isn't 0
		LCD_char(*str++);			// print out the character, then increment
	}								// the pointer down the array


void LCD_init()
	{
	LCD_PORT = 2;					// I decided to write the first command
	LCD_RS = 0;						// manually, as the busy flag can't be read
	pulse_E();						// on the first instruction
	__delay_ms(50);
	LCD_PORT = 2;
	LCD_RS = 0;
	pulse_E();
	__delay_ms(50);					// busy should be available now..
	//LCD_cmd(0x20);				// 4 bit
	LCD_cmd(0x28);					// display shift
	LCD_cmd(0x6);					// character mode
	LCD_cmd(0xc);					// display on / off and cursor
	LCD_clr();						// clear display
	}

void LCD_hex(int value)
	{
	char data;
	data = value >> 4 & 0xf;
	data = HEX_Table[data]; 				// send upper nibble
	LCD_char(data);
	data = value & 0xf;						// send lower nibble
	data = HEX_Table[data];
	LCD_char(data);
	}

void LCD_goto(char line, char column)		// combines line and lineW
	{
	unsigned char data = 0x80;				// default to 1
	if(line == 2)data = 0xc0;				// change to 2
	data += column;							// add in  column
	LCD_cmd(data);
	}

void LCD_clr()
	{
	LCD_cmd(1);								// Clr screen
	}

void LCD_cur(char on)
	{
	unsigned char cur = 0xc;				// cursor off
	if(on) cur = 0xd;						// cursor on
	LCD_cmd(cur);
	}

void LCD_busy()
	{
	unsigned char BUSY = 0;
	while(1)
		{
		LCD_TRIS = 0x0f;					// port to read mode
		LCD_RS = 0x0;
		LCD_RW = 0x1;						// set to read
		LCD_E = 0x1;
		BUSY = LCD_PORT & 0xf;				// high byte comes first
		__delay_us(5);						// so busy flag is on bit 3
		LCD_E = 0x0;
		LCD_E = 0x1;
		__delay_us(5);						// dummy read
		LCD_E = 0x0;
		LCD_RW = 0x0;						// set to write
		LCD_TRIS = 0x00;					// port to write mode
		if(!(BUSY & 0x8 )) return;			// AND with bit 3
		}
	}

void LCD_cmd(unsigned char ch)
	{
	LCD_PORT = ch >> 4 & 0xf;
	LCD_RS = 0;
	pulse_E();
	LCD_PORT = ch & 0xf;
	LCD_RS = 0;
	pulse_E();
	LCD_busy();
	}

void LCD_charD(unsigned char ch)
	{
	ch+=0x30;
	LCD_char(ch);
	}

void LCD_char(unsigned char ch)
	{
	LCD_PORT = ch >> 4 & 0xf;
	LCD_RS = 1;
	pulse_E();
	LCD_PORT = ch & 0xf;
	LCD_RS = 1;
	pulse_E();
	LCD_busy();
	}

void pulse_E()
	{
	LCD_E = 1;
	__delay_us(1);
	LCD_E = 0;
	}